"{ Encoding: utf8 }"

"
 COPYRIGHT (c) 2018 by eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
"{ Package: 'stx:libjavascript' }"

"{ NameSpace: Smalltalk }"

Object subclass:#JavaScriptCompletionEngine
	instanceVariableNames:'method class context codeView parser node requestor'
	classVariableNames:''
	poolDictionaries:''
	category:'Languages-JavaScript-Compiling & Parsing'
!

!JavaScriptCompletionEngine class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 2018 by eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
! !

!JavaScriptCompletionEngine methodsFor:'accessing'!

klass
    context notNil ifTrue:[
         ^ context receiver class
    ].
    method notNil ifTrue:[
         ^ method mclass
    ].
    class notNil ifTrue:[
         ^ class
    ].
    ^ nil

    "Created: / 23-09-2013 / 14:49:53 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!JavaScriptCompletionEngine methodsFor:'completion'!

completeForMethod:methodOrNil class:classOrNil context:contextOrNil codeView:codeViewArg into:actionBlock
    | cursorPosition |

    method := methodOrNil.
    class := classOrNil.
    context := contextOrNil.
    codeView := codeViewArg.
    requestor := actionBlock. 

    cursorPosition := codeView characterPositionOfCursor - 1.
    parser := JavaScriptCompletionParser new.
    parser ignoreErrors:true.
    parser ignoreWarnings:true.
    parser setClassToCompileFor: self klass. 
    parser nodeRewriteHook: [:n | 
        "/ Transcript showCR: (codeView textFromCharacterPosition:n startPosition to:n endPosition) asString string.
        (node isNil and:[cursorPosition between:(n startPosition) and:(n endPosition)]) ifTrue:[
             node := n.
        ].
        n
    ].
    ParseError handle:[:ex |
        "/ Transcript showCR:'JS completion: error while parsing ignored'.
        self breakPoint:#cg.
    ] do:[
        self parseSource:(codeView contents).
    ].
    
    "
    SmallSense::ParseNodeInspector openOnNode: parser tree source: codeView contents.
    "
    node notNil ifTrue:[
        |bestNode highestStart lowestEnd|
        
        "/ is there some better node?
        highestStart := node startPosition.
        lowestEnd := node endPosition.
        bestNode := node.
        node allSubNodesDo:[:each |
            ((cursorPosition between:each startPosition and:each endPosition)
              and:[ (each startPosition >= highestStart)
              and:[ (each endPosition <= lowestEnd) ]]
            ) ifTrue:[
                bestNode := each.
                highestStart := each startPosition.
                lowestEnd := each endPosition.
            ].    
        ].
        node := bestNode.
        
        node isVariableNode ifTrue:[
             self completeVariable.
             ^ self
        ].
        node isMessage ifTrue:[
             self completeMessage.
             ^ self
        ].
    ].

    self breakPoint: #jv.

    "Created: / 23-09-2013 / 13:49:56 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified (format): / 25-09-2013 / 15:07:19 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified (format): / 04-06-2019 / 16:20:43 / Claus Gittinger"
!

parseSource:source 
    "the actual call to the parser; redefinable for JS-like languages
     which may have more than a single method/function to parse"
     
    parser parseMethod:source in:(self klass) ignoreErrors:true ignoreWarnings:true.

    "Created: / 04-06-2019 / 16:14:20 / Claus Gittinger"
! !

!JavaScriptCompletionEngine methodsFor:'private-completion helpers'!

completeGlobalVariables
    | suggestedVariables editAction klass namespace |

    klass := self klass.
    
    suggestedVariables := OrderedCollection new.

    "/ First, globals starting with typed prefix
    namespace := klass nameSpace.
    [ namespace isNameSpace ] whileFalse:[ namespace := namespace nameSpace ]. "/#nameSpace may return private class, sigh
    namespace keysDo:[:name|
        (name startsWith: node name) ifTrue:[
            suggestedVariables add: name.                    
        ].
    ].
    Smalltalk keysDo:[:name|
        (name startsWith: node name) ifTrue:[
            suggestedVariables add: name.                    
        ].
    ].
    "/ Ignore class variables, pools and private classes, probably
    "/ useless. We will see.

    editAction := self editorToCompleteNode:node in:codeView fromSelectionIn:suggestedVariables.
    requestor value:suggestedVariables value:editAction value:'Variable'

    "Created: / 25-09-2013 / 10:48:43 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified (format): / 14-02-2019 / 10:59:32 / Claus Gittinger"
!

completeLocalVariables
    |klass instvars variables variables1 variables2 editor|

    klass := self klass.
    klass notNil ifTrue:[
        instvars := klass allInstVarNames
    ] ifFalse:[
        instvars := #()
    ].
    variables := parser methodArgs , parser methodVars , instvars 
            , #( 'this' ).
     
    "/ Make those starting with typed prefix first in the list.
    
    variables1 := (variables select:[:name | name startsWith:node name ]) sort.
    variables2 := (variables reject:[:name | name startsWith:node name ]) sort.
    variables := variables1 , variables2.
    editor := self 
            editorToCompleteNode:node
            in:codeView
            fromSelectionIn:variables.
    requestor 
        value:variables
        value:editor
        value:'Variable'

    "Created: / 04-06-2019 / 16:52:17 / Claus Gittinger"
!

completeMessage
    | already suggestions selector receiver receiverClass editor|

    selector := node selector.
    receiver := node receiver.
    receiverClass := self classOfNode:receiver.

    "/ I hate this mirror/non-mirror confusion - a q&d hack
    (receiverClass isMemberOf:(Smalltalk at:#Workflow::CodeCompletionClassMirror)) ifTrue:[
        receiverClass := receiverClass reflectedClass
    ].
    suggestions  := Parser 
                        findBest:15 selectorsFor:selector 
                        in:receiverClass forCompletion:true
                        ignoreIfAnnotatedWith:#noCodeCompletionInExpecco.

    already := Set new.
    
    "/ Now, have to translate selectors from Smalltalk to JavaScript form
    suggestions := suggestions 
                        collect:[:eachSTSelector |
                            |suggestion jsSelector restLine|

                            jsSelector := self selectorToFunctionName:eachSTSelector.
                            "/ append parentheses, but only if there is not already one
                            restLine := codeView lineStringFromCursor withoutSeparators.
                            suggestion := jsSelector.    
                            (restLine isEmpty or:[restLine startsWith:')']) ifTrue:[
                                eachSTSelector numArgs > 0 ifTrue:[
                                    suggestion := jsSelector,'(...)' 
                                ] ifFalse:[    
                                    suggestion := jsSelector,'()'   
                                ].
                            ].
                            (already includes:suggestion) ifTrue:[
                                nil
                            ] ifFalse:[
                                already add:suggestion.
                                suggestion
                            ].
                        ]
                        thenSelect:[:eachSTSelectorOrNil | eachSTSelectorOrNil notNil].

    editor := [:index |
        |answer selPos start stop oldText oldLen newLen |

        answer := suggestions at:index.

        selPos := node selectorPosition.
        selPos notNil ifTrue:[ 
            start := selPos first.
            stop := selPos last.
        ] ifFalse:[ 
            start := node endPosition - selector size + 1.
            stop := node endPosition.
        ].
        oldText := (codeView textFromCharacterPosition:start to:stop) asString string withoutSeparators.

        oldLen := stop - start + 1.
        newLen := answer size.

        codeView
            undoableDo:[ 
                codeView replaceFromCharacterPosition:start to:stop with:(answer).

                (answer startsWith:oldText) ifTrue:[
                    codeView selectFromCharacterPosition:start+oldLen to:start+newLen-1.
                ] ifFalse:[
                    codeView selectFromCharacterPosition:start to:start+newLen-1.
                ].
                codeView dontReplaceSelectionOnInput
            ]
            info:'Completion'.
        ].

    requestor value: suggestions value: editor value: 'Function/Method'

    "Created: / 25-09-2013 / 11:36:26 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 17-11-2014 / 13:45:49 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 04-06-2019 / 17:20:39 / Claus Gittinger"
!

completeVariable

    node name first isUppercase ifTrue:[
         self completeGlobalVariables
    ] ifFalse:[
         self completeLocalVariables
    ].

    "Created: / 23-09-2013 / 21:05:28 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified (format): / 25-09-2013 / 10:48:50 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 04-06-2019 / 16:53:27 / Claus Gittinger"
! !

!JavaScriptCompletionEngine methodsFor:'private-helpers'!

classOfNode:aNode
    "when showing possible completions for a message, 
     it is a good idea to know what the reveiver is.
     This method tries to ficure out what class the node might have
     (mostly heuristics)"

    | nm nodeVal receiverClass nodeSelector nodeReceiver klass |

    klass := self klass.

    aNode isLiteral ifTrue:[
        ^ aNode value class
    ].

    aNode isThis ifTrue:[
        ^ klass ? UndefinedObject
    ].

    aNode isVariable ifTrue:[
        nm := aNode name.
        nm = 'super' ifTrue:[
            klass isNil ifTrue:[^ Object].
            ^ klass superclass
        ].
        nm isUppercaseFirst ifTrue:[
            | namespace global nmAsSymbol |

            nmAsSymbol := nm asSymbolIfInterned.
            nmAsSymbol notNil ifTrue:[
                klass notNil ifTrue:[
                    namespace := klass nameSpace.
                ] ifFalse:[
                    namespace := Smalltalk
                ].    
                [ namespace isNameSpace ] whileFalse:[ namespace := namespace nameSpace ]. "/#nameSpace may return private class, sigh
                global := namespace at: nmAsSymbol.
                global notNil ifTrue:[ ^ global class ].
                global := Smalltalk at: nmAsSymbol.
                global notNil ifTrue:[ ^ global class ].
            ].

            (klass notNil and:[klass theNonMetaclass classVarNames includes:aNode name]) ifTrue:[
                nodeVal := klass theNonMetaclass classVarAt:aNode name.
                ^ nodeVal class.
            ].
        ] ifFalse:[
            context notNil ifTrue:[
                "/ a local in the context?
                (context argAndVarNames includes:nm) ifTrue:[
                    nodeVal := context argsAndVars at:(context argAndVarNames indexOf:nm).
                    nodeVal notNil ifTrue:[
                        ^ nodeVal class
                    ].
                ]
            ].

            klass notNil ifTrue:[
                (klass allInstVarNames includes:nm) ifTrue:[
                    | idx |

                    idx := klass instVarIndexFor: nm.
                    [
                        ^ (klass anyInstance instVarAt: idx) class
                    ] on: Error do:[:ex|

                    ].
                ].
            ]
        ].
    ].

    aNode isMessage ifTrue:[
        nodeSelector := aNode selector.
        nodeReceiver := aNode receiver.

        "/ some hardwired knowledge here
        ( #( #'new' #'basicNew' #'new:' #'basicNew:' #'with:' #'with:with:') includes: nodeSelector ) ifTrue:[
            receiverClass := self classOfNode:nodeReceiver.
            receiverClass notNil ifTrue:[
                receiverClass isBehavior ifTrue:[
                    receiverClass isMeta ifTrue:[
                        ^ receiverClass theNonMetaclass
                    ]
                ]
            ].
        ].
        klass notNil ifTrue:[
            (nodeReceiver isSelf and:[nodeSelector = #'class']) ifTrue:[
                ^ klass class
            ].
        ].

        (nodeSelector = #'asFilename') ifTrue:[
            ^ Filename
        ].
        (nodeSelector = #'size') ifTrue:[
            ^ SmallInteger
        ].

        "/ some wellknown boolean returners (need better type inference here)
        (#( isNil notNil not isEmptyOrNil notEmptyOrNil notEmpty isEmpty 
            isBehavior isMeta
            = ~= == ~~ > >= < <= between:and:
            includes: contains:
            and: or:
            exists atEnd
        ) includes:nodeSelector ) ifTrue:[
            ^ True "/ not Boolean, because it does not contain the full protocol (would not find ifTrue:)
        ].

        ( #( + - * / // \\ ) includes:nodeSelector) ifTrue:[
            "/ assume numeric
            ^ Number
        ].
        ( #( asFloat ) includes:nodeSelector) ifTrue:[
            ^ Float
        ].

        ( #( class theMetaclass theNonMetaclass ) includes:nodeSelector) ifTrue:[
            "/ assume behavior
            ^ Behavior
        ].
    ].
    ^ Object

    "Created: / 25-09-2013 / 14:17:14 / Jan Vrany <jan.vrany@fit.cvut.cz>"
    "Modified: / 03-03-2017 / 17:09:49 / cg"
    "Modified (comment): / 04-06-2019 / 16:55:48 / Claus Gittinger"
!

editorToCompleteNode:node in:codeView fromSelectionIn:variables
    "private helper: return an editAction to replace
     the word before the cursor by the selected variable from variables.
     variables being a list of suggested completions."
     
    ^ [:index |
        |answer start stop oldVar oldLen newLen |

        answer := variables at:index.

        start := node startPosition.
        stop := node endPosition.
        oldVar := (codeView textFromCharacterPosition:start to:stop) asString string withoutSeparators.

        oldLen := stop - start + 1.
        newLen := answer size.

        codeView
            undoableDo:[ 
                codeView replaceFromCharacterPosition:start to:stop with:(answer).

                (answer startsWith:oldVar) ifTrue:[
                    codeView selectFromCharacterPosition:start+oldLen to:start+newLen-1.
                ] ifFalse:[
                    codeView selectFromCharacterPosition:start to:start+newLen-1.
                ].
                codeView dontReplaceSelectionOnInput
            ]
            info:'Completion'.
    ].

    "Created: / 14-02-2019 / 10:53:39 / Claus Gittinger"
!

selectorToFunctionName: selector
    | numArgs |

    numArgs := selector numArgs.
    numArgs == 0 ifTrue:[
         ^ selector 
    ].
    numArgs == 1 ifTrue:[
        selector isBinarySelector ifTrue:[
             ^ selector 
             "/ self error: 'Invalid JavaScript function/method name'.
        ].
        ^ selector copyButLast: 1. 
    ].
    ^ (selector copyButLast:1) replaceAll: $: with: $_.

    "Created: / 25-09-2013 / 13:45:14 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!JavaScriptCompletionEngine class methodsFor:'documentation'!

version
    ^ '$Header$'
!

version_CVS
    ^ '$Header$'
! !

